<?php

namespace Bydedi\Database;

use Bydedi\Cache\CacheInterface;
use Exception;
use mysqli;

/**
 * Class Connection
 */
class MysqlConnection implements ConnectionInterface
{

    const SQL_DEBUG = false;

    const SQL_NONE = 1;
    const SQL_ALL = 2;
    const SQL_INIT = 3;

    const SQL_ASSOC = 1;
    const SQL_INDEX = 2;

    /**
     * @var array
     */
    protected $structures = array();

    /**
     * @var \mysqli
     */
    public $db = null;
    public $result = null;
    public $record = null;
    public $total = null;

    /**
     * @var CacheInterface
     */
    protected $cache;

    /**
     * @param CacheInterface $cache
     */
    public function __construct(CacheInterface $cache)
    {
        $this->cache = $cache;
        $this->connect();
    }

    /**
     * load the table structures from the cache
     *
     * @return void
     */
    public function loadStructures()
    {
        $structCache = $this->cache->get('structures');
        if ($structCache !== false) {
            $this->structures = $structCache;
        }
    }

    /**
     * get the $structures property
     *
     * @return array
     */
    public function getStructures()
    {
        return $this->structures;
    }

    /**
     * init the connection
     *
     * @return void
     */
    public function connect()
    {
        $this->db = new mysqli(
            getenv('DB_HOST'),
            getenv('DB_USERNAME'),
            getenv('DB_PASSWORD'),
            getenv('DB_DATABASE'),
            getenv('DB_PORT') ?: 3306
        );

        if ($this->db->connect_error) {
            die('Erreur de connexion (' . $this->db->connect_errno . ') ' . $this->db->connect_error);
        }

        $this->db->set_charset("utf8");
    }

    /**
     * close the connection
     *
     * @return bool
     */
    public function disconnect()
    {
        return $this->db->close();
    }

    /**
     * execute a statement SQL
     *
     * @param string $sql    the statement SQL
     * @param int    $type   type of result
     * @param int    $format format of array to return
     *
     * @return bool
     */
    public function query($sql, $type = self::SQL_NONE, $format = self::SQL_INDEX)
    {

        /*$dbg = Application::getDebugBar();
        if (getenv('APP_DEBUG') && !empty($dbg)) {

            $sqlLog = substr($sql, 0, 150);
            if ($sqlLog != $sql) {
                $sqlLog.= '...';
            }

            $dbg['time']->startMeasure('statement', $sqlLog);
        }*/

        $this->db->query("SET SQL_BIG_SELECTS=1");

        $this->record = array();

        $this->result = $this->db->query($sql);

        if ($this->db->error) {
            error_log($sql);
            error_log($this->db->errno . ' : ' . $this->db->error);
            return false;
        }

        if ($this->result !== false && $this->result instanceof \mysqli_result) {
            $resulttype = ($format == self::SQL_INDEX) ? MYSQLI_NUM : MYSQLI_ASSOC;
            if ($type == self::SQL_ALL) {
                while ($row = $this->result->fetch_array($resulttype)) {
                    $this->record[] = $row;
                }
            } else {
                $this->record = $this->result->fetch_array($resulttype);
            }
            $this->result->free();
        }
        return true;
    }

    /**
     * Execute a SQL request and return all result
     *
     * @param string $sql the SQL request to execute
     *
     * @return array
     */
    public function getAll($sql)
    {
        $this->query($sql, self::SQL_ALL, self::SQL_ASSOC);
        return $this->record;
    }

    /**
     * Execute a SQL request and return the first result
     *
     * @param string $sql the SQL request to execute
     *
     * @return array
     */
    public function getOne($sql)
    {
        $this->query($sql, self::SQL_INIT, self::SQL_ASSOC);
        return $this->record;
    }

    /**
     * return the primary key of the table given
     *
     * @param string $table name of the table
     *
     * @return bool
     */
    public function getPrimaryKey($table)
    {
        $result = $this->getStructure($table);
        foreach ($result as $r) {
            if ($r['Key'] == 'PRI') {
                return $r['Field'];
            }
        }

        return false;
    }

    /**
     * return the struct of the table given
     *
     * @param string $table name of the table
     *
     * @return array
     */
    public function getStructure($table)
    {

        if (empty($this->structures[$table])) {
            $this->structures[$table] = $this->getAll("DESC " . $table);
        }

        return $this->structures[$table];
    }

    /**
     * Fonction pour sauvegarder les données du tableau $data dans la table $table
     *
     * @param string $table nom de la table dans laquelle il faut insérer des data
     * @param array  $data  tableau de donées a insérer
     *
     * @return mixed
     * @throws Exception
     */
    public function save($table, $data)
    {

        $pk = $this->getPrimaryKey($table);

        if (empty($pk)) {
            throw new Exception('no primary key in table '. $table);
        }

        $autofill = array(
            $table . '_date_ajout' => 'NOW()',
            $table . '_date_modif' => 'NOW()',
        );

        $struct = $this->getStructure($table);

        $q = new Query();
        $action = (!empty($data[$pk])) ? 'update' : 'insert';
        $q->$action($table);

        $this->escapeArray($data);

        foreach ($struct as $field) {
            if (in_array($field['Field'], array_keys($autofill))) {
                if ($field['Field'] == $table . '_date_ajout') {
                    if ($action == 'insert') {
                        $q->value($field['Field'], $autofill[$field['Field']]);
                    }
                } else {
                    $q->value($field['Field'], $autofill[$field['Field']]);
                }
            }

            if (!isset($data[$field['Field']])) {
                continue;
            }

            $q->value($field['Field'], $data[$field['Field']]);
        }

        if ($action == 'update') {
            $q->where($pk, $data[$pk]);
        }

        $this->query($q->build());

        if ($action == 'insert') {
            $data[$pk] = $this->db->insert_id;
        }

        return $data[$pk];
    }

    /**
     * @param $value
     * @return string
     */
    public function escape($value)
    {
        return $this->db->real_escape_string($value);
    }

    public function escapeArray(&$vars, $key = '')
    {
        if (is_array($vars)) {
            foreach ($vars as $key => &$value) {
                $this->escapeArray($value, $key);
            }
        } else {
            $vars = $this->escape($vars);
        }
    }

    /**
     * Démarre une nouvelle transaction
     *
     * @return void
     */
    public function beginTransaction()
    {
        $this->db->autocommit(false);
    }

    /**
     * Fin de la transaction
     *
     * @return bool
     */
    public function endTransaction()
    {
        $result = $this->db->commit();
        $this->db->autocommit(true);

        return $result;
    }
}
